#! wish -f        -*- tcl -*-

#
# This file was written by
#         Donal K. Fellows (fellowsd@cs.man.ac.uk)
# on
#         3rd April 1994
# in response to a request from someone on comp.lang.tcl about how to put context sensitive help  la
# WinDoze on Tcl/Tk menus.  As the idea interested me, I had a go at implementing this, and I came up with
# this...
#
#
# You may distribute and modify this file however you like, provided you acknowledge me as the
# original author and include this notice.  Donations to the Donal K. Fellows Beverage Consumption
# Support Fund are not necessary, even if they are recommended (by me! :)
#
#
# If there are bugs in this, and these haven't been introduced by someone else, mail me and I'll
# probably try to fix them, though I'm not making any promises...  Alternatively, as you have The
# Force, Use It... :) :)
#
#
# Anyone accusing my code of being idiomatic Tcl is probably more than a little misguided...
#

### -------------------------------------------------------------------------------------------------------
#
# PROCEDURE
#   SuperMenu
#
# DESCRIPTION
#   Creates a menu bar with menus (no nested menus as yet) with context-sensitive help.
#   DOES NOT PACK MENU THE MENUBAR (but does handle all components)
#
# ARGUMENTS
#   menubar              The path name of the menu-bar.                         NOT A KEYED ARGUMENT.
#   helpvar              Name of variable containing current c-s help message.  NOT A KEYED ARGUMENT.
#   menus                List of menubutton definitions
#
# Each menubutton is as normal except:
#   Omit the `menubutton pathname' bit (supplied by this procedure)
#   Accepts the following special keyed args:
#       -help            Help message for the menu button
#       -contents        Menu definition for menu hanging off menu button
#       -right           Packs this menu to the right
#       -tearoff         Passed in to the menu generator
#   Menubuttons have pathnames consisting of a number appended to the menu-bar pathname (e.g. .menu.2)
#
# Each menu is as normal except:
#   Omit the widget creation (will have a name like .menu.2.menu)
#   Include a list of things to add to the menu, each time omitting the `pathname add' (supplied by proc.)
#   Each item added takes an extra keyed argument:
#       -help            Help message for menu item
#       -accel           Accelerator for the menu item
#   With a cascade entry, instead of specifying the menu name, specify a definition of the menu in the same
#   style as has just been outlined here.
#
# REQUIRES
#   parseargs            option parser
#
# GLOBAL VARIALBES
#   supermenu(help,...)  array of context help messages (can support multiple simultaneous menu bars)
#   supermenu(olditem)   used to reduce the number of updates of the screen
#   supermenu_bindname   array used to translate keysyms to characters
#
# (SEMI-)PUBLIC ROUTINES
#   supermenu_cvtaccel     Translates a binding into an accelerator string
#   supermenu_fixarray     Allows (initially) sparse arrays
#   supermenu_leaveaction  What happens when you leave a menu or menubutton
#   supermenu_mkmenu       Builds a menu from spec, possibly with a recursive call to itself...
#   supermenu_motionaction The really clever bit; handles translation of motion in menus into help messages
#
# BUGS/DEFICIENCIES
#   Still a bit simple to have many bugs... :)
#   The accelerator binding parser is very stupid, and only really understands letters and
#     letters with modifiers. (?)
#
# TO DO
#   1) Allow cascaded menus.
#          DONE
#   2) Handle multiple right-anchored menus on the menubar better.
#   3) Read definitions/help from a file?
#          NO, APP WRITER CAN DO THAT FOR THEMSELVES
#   4) Handle %^& etc.
#          DONE
#   5) Handle keyboard traversal.
#

proc SuperMenu {menubar helpvar details} {
    catch {
	frame $menubar -relief raised -bd 2
    }
    set bindwidget [winfo toplevel $menubar]
    set count 0
    set parseable(-help)     1
    set parseable(-contents) 1
    set parseable(-right)    0
    set parseable(-tearoff)  1
    set menuparseable(-help) 1
    foreach mbaritem $details {
	incr count
	eval [list menubutton $menubar.$count -menu $menubar.$count.menu ] \
		[parseargs parseable parsed $mbaritem]
	eval [list menu $menubar.$count.menu ] \
		[if [string length $parsed(-tearoff)] {list -tearoff $parsed(-tearoff)}]
	if ($parsed(-right)) {
	    pack $menubar.$count -side right
	} else {
	    pack $menubar.$count -side left
	}
	bind $menubar.$count      <Enter>  [list +set $helpvar $parsed(-help)]
	bind $menubar.$count      <Leave>  [list +supermenu_leaveaction $helpvar]
	supermenu_mkmenu $menubar.$count.menu $helpvar $parsed(-contents) $bindwidget
    }
}

# --- BEGINNING OF PRIVATE PROCEDURES --------------------------------------------------------------------

;proc supermenu_mkmenu {menuname helpvar contents bindwidget} {
    global supermenu
    set menuparseable(-help)  1
    set menuparseable(-menu)  1
    set menuparseable(-accel) 1

    if ![winfo exists $menuname] {
	eval menu $menuname -tearoff 0		;# Never create tearoff menus!
    }

    bind $menuname <Leave>  [list +supermenu_leaveaction $helpvar]
    bind $menuname <Motion> [list +supermenu_motionaction $menuname $helpvar]
    
    set mitem [$menuname cget -tearoff]
    foreach menuItem $contents {
	set menuargs [parseargs menuparseable special $menuItem]
	if [string match cascade [lindex $menuargs 0]] {
	    lappend menuargs -menu $menuname.${mitem}menu
	    supermenu_mkmenu $menuname.${mitem}menu $helpvar $special(-menu) $bindwidget
	}
	if [string length $special(-accel)] {
	    lappend menuargs -accelerator [supermenu_cvtaccel $special(-accel)]
	}
	eval $menuname add $menuargs
	if [string length $special(-accel)] {
	    bind $bindwidget $special(-accel) "+$menuname invoke $mitem"
	}
	set chi "help,$menuname,$mitem"
	set supermenu($chi) $special(-help)
	incr mitem
    }
}

# Lookup table for 
set supermenu_bindname(exclam)		{!}
set supermenu_bindname(at)		{@}
set supermenu_bindname(numbersign)	{#}
set supermenu_bindname(dollar)		{$}
set supermenu_bindname(percent)		{%}
set supermenu_bindname(asciicircum)	{^}
set supermenu_bindname(ampersand)	{&}
set supermenu_bindname(asterisk)	{*}
set supermenu_bindname(parenleft)	{(}
set supermenu_bindname(parenright)	{)}
set supermenu_bindname(minus)		{-}
set supermenu_bindname(equal)		{=}
set supermenu_bindname(backslash)	{\\}
set supermenu_bindname(grave)		{`}
set supermenu_bindname(underscore)	{_}
set supermenu_bindname(plus)		{+}
set supermenu_bindname(bar)		{|}
set supermenu_bindname(asciitilde)	{~}
set supermenu_bindname(semicolon)	{;}
set supermenu_bindname(comma)		{,}
set supermenu_bindname(period)		{.}
set supermenu_bindname(slash)		{/}
set supermenu_bindname(colon)		{:}
set supermenu_bindname(less)		{<}
set supermenu_bindname(greater)		{>}
set supermenu_bindname(question)	{?}
set supermenu_bindname(braceleft)	{\{}
set supermenu_bindname(braceright)	{\}}
set supermenu_bindname(quotedbl)	{\"}
set supermenu_bindname(bracketleft)	{\[}
set supermenu_bindname(bracketright)	{\]}
set supermenu_bindname(apostrophe)	{'}

# Still really needs a bit more work (only handles one <> binding _or_ a character string)
;proc supermenu_cvtaccel accelbind {
    global supermenu_bindname

    set accellist {{} {} {} {}}
    set ab $accelbind
    while {[string length $ab]} {
	if {[string range $ab 0 0] == " "} {
	    set ab [string range $ab 1 end]
	    continue
	}
	if [regexp {<([^<>]+)>(.*)} $ab wibble bound ab] {
	    # Is a complex binding
	    set blst {}
	    foreach bnd [split $bound -] {
		switch -regexp $bnd {
		    ^[A-Z]$ {
			lappend blst Shft [string toupper $bnd]
		    }
		    ^.$ {
			lappend blst [string toupper $bnd]
		    }
		    Control {
			lappend blst Ctrl
		    }
		    Shift {
			lappend blst Shft
		    }
		    Lock  -
		    Alt   -
		    Meta {
			lappend blst $bnd
		    }
		    Key[Pp]ress   -
		    Key[Rr]elease -
		    Key {
			# Ignore
		    }
		    .+ {
			if [catch {
			    lappend blst $supermenu_bindname($bnd)
			}] {
			    # Don't know what to do, so capitalise the first letter, and append :^)
			    lappend blst [string toupper [string range $bnd 0 0]][string tolower \
				    [string range $bnd 1 end]]
			}
		    }
		    default {
			# Everyone think about why this works...
			lappend blst -
		    }
		}
	    }
	    lappend accellist [join $blst +]
	    continue
	}
	# Is a simple binding
	regexp {(.)(.*)} $ab wibble chr ab
	if [regexp {[A-Z]} $chr] {
	    lappend accellist "Shft+$chr"
	} else {
	    lappend accellist [string toupper $chr]
	}
    }
    return [join $accellist " "]
}

;proc supermenu_motionaction {menuname helpvar} {
    upvar   $helpvar h
    global  supermenu

    set active [$menuname index active]
    if {$active == "none"} {
	set supermenu(olditem) {}
	set h {}
	return
    }
    set newCHI "help,$menuname,$active"
    if [string compare $newCHI $supermenu(olditem)] {
	set supermenu(olditem) $newCHI
	set h $supermenu($newCHI)
    }
}

;proc supermenu_leaveaction {helpvar} {
    upvar   $helpvar h
    global  supermenu
    set h {}
    set supermenu(olditem) {}
}

;proc supermenu_fixarray {n e o} {
    upvar  $n x
    if {[lsearch -exact [array names x] $e] == -1} {
	set x($e) ""
    }
}
set supermenu(olditem) {}
trace variable supermenu r supermenu_fixarray
